module types
implicit none
integer, parameter  ::  i4b=selected_int_kind(9)
integer, parameter  ::  dp=kind(1.0d0)
end module types

module globals
use types
implicit none
integer(i4b) :: lmax,rlim,k,kp,r,rp,l,m,rmax,rmaxofk,info,sigma
real(dp) :: lambda,U,eps,V,temp,alambda
real(dp) :: tempold,entold,noimp
real(dp) :: elemaddedsite(-1:1,-1:2,-1:2)
real(dp),allocatable :: tn(:),epsn(:),h(:,:),work(:),energies(:)
real(dp),allocatable :: elem(:,:,:)
character(len=10) :: ch
end module globals

!*********************!
!*********************!
!*********************!

program nrg
use types
use globals
implicit none

open(1,file='input.dat')
read(1,*) lmax,ch
read(1,*) rlim,ch
read(1,*) U,ch
read(1,*) eps,ch
read(1,*) V,ch
close(1)
lambda=3.0d0
open(2,file='entropy.dat')
open(3,file='thermo_noimp.dat')

! Allocate memory for arrays:
allocate (tn(-1:lmax),epsn(-1:lmax),work(1:200*rlim))


! Define states of added site, N+1:
! |N+1;k=0>  = |->           = |vac>
! |N+1;k=+1> = |up>          = c+_{up}|vac>
! |N+1;k=-1> = |down>        = c+_{down}|vac>
! |N+1;k=2>  = |up and down> = c+_{up}c+_{down}|vac>

! Define matrix product state of full system
! |N+1;k,r> = |N+1;k> X |N;r>
! where |N+1;r> = Sum_{k,r'} U_{N+1}^r(k,r') |N+1;k,r'> are the diagonalized states at iteration N+1.

! Matrix elements between states of added site:
! elemaddedsite(x,y,z) = <N+1;k=y| c_{x} |N+1;k=z>, with x=1 for up and x=-1 for down, FOR ANY ITERATION N.
elemaddedsite=0.0d0
elemaddedsite(+1,0,+1) = +1d0
elemaddedsite(+1,-1,2) = +1d0
elemaddedsite(-1,0,-1) = +1d0
elemaddedsite(-1,+1,2) = -1d0


! Wilson chain parameters
do l=0,lmax
tn(l)=0.5d0*(1.0d0+lambda**(-1))*(1.0d0-lambda**(-l-1))*((1.0d0-lambda**(-2*l-1))*&
(1.0d0-lambda**(-2*l-3)))**(-0.5d0)
epsn(l)=0.0d0
end do
alambda=(lambda+1.0d0)/(lambda-1.0d0)*log(lambda)/2.0d0


! Setup initial Hamiltonian:
write(6,*) "Anderson impurity model"
write(6,*) "Lambda= ",lambda
write(6,*) "max number of states kept at each iteration= ",rlim
call initialham


! Main iteration loop:
do l=0,lmax
write(6,'(x,a,i3)') "Iteration: ", l
rmaxofk=rmax
rmax=4*rmax


! Construct Hamiltonian matrix:
allocate (h(1:rmax,1:rmax))
call ham
deallocate(elem,energies)
allocate(energies(1:rmax))


! Diagonalize Hamiltonian matrix:
call dsyev('V','U',rmax,h,rmax,energies,work,200*rlim,info)

! Energies relative to ground state energies:
energies(:)=energies(:)-energies(1)

! Calculate thermodynamics at this iteration
call thermo


! Truncation of states:
rmax=min(rlim,rmax)
write(6,*) "States kept at this iteration=",rmax
write(6,*) " "


! Calculate matrix elements needed for the construction of Hamiltonian at next iteration:
allocate(elem(-1:1,1:rmax,1:rmax))
call elemcalc
deallocate(h)


! End main iteration loop
end do !l
write(6,*) " "
write(6,*) "Calculation Complete!"

end program nrg



!*********************!
!*********************!  Subroutines:
!*********************!



subroutine ham
use types
use globals
implicit none


h=0.0d0
! Off-diagonal elements of the Hamiltonian
do k=-1,2
do kp=-1,2

! Hop an up-spin (sigma=+1) or down-spin (sigma=-1) electron from site (l-1) to site l:
do sigma=-1,1,2
if (elemaddedsite(sigma,k,kp).ne.0) then
do r=1,rmaxofk
do rp=1,rmaxofk
h(rmaxofk*(k+1)+r,rmaxofk*(kp+1)+rp) = h(rmaxofk*(k+1)+r,rmaxofk*(kp+1)+rp) + &
tn(l-1)*elem(sigma,r,rp)*elemaddedsite(sigma,k,kp)*(-1.0d0)**k
end do !rp
end do !r
end if
end do

end do !kp
end do !k

! Hermitian conjugate
h(:,:)=h(:,:)+transpose(h(:,:))


! Diagonal elements of the Hamiltonian
m=0
do k=-1,2
   do r=1,rmaxofk
      m=m+1
      h(m,m)=sqrt(lambda)*energies(r)+abs(k)*epsn(l)
   end do
end do


end subroutine ham



!*********************!
!*********************!
!*********************!



subroutine elemcalc
use types
use globals
implicit none

! matrix elements of c+_{l,sigma}, where sigma=-1 is down-spin and sigma=+1 is up-spin.
! NOTE: only calculate matrix elements between KEPT states.

elem=0.0d0
do sigma=-1,1,2 

do k=-1,2
do kp=-1,2
if (elemaddedsite(sigma,kp,k).eq.0) cycle

call dgemm('T','N',rmax,rmax,rmaxofk,elemaddedsite(sigma,kp,k),&
h(rmaxofk*(k+1)+1:rmaxofk*(k+2),1:rmax),rmaxofk,&
h(rmaxofk*(kp+1)+1:rmaxofk*(kp+2),1:rmax),rmaxofk,&
1.0d0,elem(sigma,1:rmax,1:rmax),rmax)

end do
end do

end do


end subroutine elemcalc



!*********************!
!*********************!
!*********************!



subroutine thermo
use types
use globals
implicit none
real(dp) :: pf,bf,entropy,betabar

! Calculate some thermodynamics:
betabar=0.9d0
temp=lambda**(-0.5d0*(l-1))/betabar
write(6,'(x,a,2x,e13.6)') "Temperature= ",temp
pf=0
entropy=0
do r=1,rmax
   bf=exp(-betabar*energies(r))    ! Boltzmann factor
   pf=pf+bf                        ! Partition function
   entropy=entropy+betabar*energies(r)*bf 
end do
entropy=entropy/pf+log(pf)
read(3,*) noimp
entropy=entropy-noimp

    
if (l.gt.0) then
write(2,'(e21.14,1x,e21.14)') sqrt(temp*tempold),(entropy+entold)/2.0d0
flush(2)
end if
tempold=temp
entold=entropy

end subroutine thermo



!*********************!
!*********************!
!*********************!



subroutine initialham
use types
use globals
implicit none

write(6,*) "U/D = ",U
write(6,*) "epsilon/D = ",eps
write(6,*) "V/D = ",V
write(6,*) " "

! Hopping between impurity and Wilson chain 'zero orbital' is V (not tn).
! Due to rescaling of Hamiltonian at each step, need to rescale parameters of H_0 also!
tn(-1)=sqrt(alambda/lambda)*V
epsn(-1)=0.0d0
eps=eps/lambda
U=U/lambda

! Dimensionality of impurity Hilbert space:
rmax=4

! Energies of impurity states
allocate(energies(1:rmax))
energies(1)=eps
energies(2)=0.0d0
energies(3)=eps
energies(4)=2.0d0*eps + U

! Matrix elements between impurity states is in this case as defined earlier (in fact, for their H.c.)
allocate(elem(-1:1,1:4,1:4))
elem(-1,1:4,1:4)=transpose(elemaddedsite(-1,-1:2,-1:2))
elem(+1,1:4,1:4)=transpose(elemaddedsite(+1,-1:2,-1:2))


end subroutine initialham



!*********************!
!*********************! End of program
!*********************!
